/* CS 106X Lecture 8: Recursion
 * -----------------
 * This program compares the runtimes of
 * a non-memoized and memoized Fibonacci
 * implementation.  Note that some variables
 * are of type "long" since those have more space
 * than ints to store numbers, and Fibonacci numbers
 * get very large very quickly!
 */

#include <iostream>
#include "console.h"
#include "hashmap.h"
#include "simpio.h"
#include "strlib.h"

using namespace std;

void fibNormal();
void fibMemoize();
long fibonacci(int i, HashMap<int, long>& cache);
long fibonacci(int i);

const int MAX_FIB_INDEX = 75;

int main() {
    string response = getLine("[N]ormal or [M]emoized? ");
    if (startsWith(response, "N")) {
        fibNormal();
    } else if (startsWith(response, "M")) {
        fibMemoize();
    }

    return 0;
}

/**
 * Function: fibNormal
 * -------------------
 * Calculates the first MAX_FIB_INDEX fibonacci numbers
 * using the non-memoized Fibonacci function.
 */
void fibNormal() {
    for (int i = 0; i < MAX_FIB_INDEX; i++) {
        long fibNum = fibonacci(i);
        cout << "Fibonacci number " << (i+1) << " is " << fibNum << endl;
    }
}

/**
 * Function: fibMemoize
 * -------------------
 * Calculates the first MAX_FIB_INDEX fibonacci numbers
 * using the memoized Fibonacci function.
 */
void fibMemoize() {
    for (int i = 0; i < MAX_FIB_INDEX; i++) {
        HashMap<int, long> cache;
        long fibNum = fibonacci(i, cache);
        cout << "Fibonacci number " << (i+1) << " is " << fibNum << endl;
    }
}

/**
 * Function: fibonacci
 * -------------------
 * Calculates the i'th Fibonacci number, where i
 * is 0 indexed.  Takes exponential time.
 */
long fibonacci(int i) {
    if (i < 0) {
        throw "Illegal negative index";
    } else if (i < 2) {
        return i;
    } else {
        return fibonacci(i-1) + fibonacci(i-2);
    }
}

/**
 * Function: fibonacci (memoized)
 * -------------------
 * Calculates the i'th Fibonacci number, where i
 * is 0 indexed.  Requires a map to be passed in
 * of previously-calculated Fibonacci numbers
 * (initially empty).  Takes linear time.
 */
long fibonacci(int i, HashMap<int, long>& cache) {
    if (i < 0) {
        throw "Illegal negative index";
    } else if (i < 2) {
        return i;
    } else if (cache.containsKey(i)) {
        return cache[i];
    } else {
        long result = fibonacci(i-1, cache) + fibonacci(i-2, cache);
        cache[i] = result;
        return result;
    }
}



